#include "Compiler.h"
#include "codeGen/Quadruple.h"
#include "codeGen/RegisterAllocator.h"
#include <filesystem>
#include <iostream>
#include <cassert>

#define COMMENT "//"

extern std::filesystem::path binaryPath;
const std::string Compiler::extension = "s";
void Compiler::externalCommand(std::filesystem::path lat) {
	auto obj = lat;
	obj.replace_extension("o");
	auto cmd = std::string("as --32 '") + lat.string() + "' -o '" + obj.string() + "'";
	cerr << cmd << endl;
	system(cmd.data());
	auto lib = binaryPath.parent_path() / "lib";
	auto runtime = lib / "runtime.o";
	auto libc = lib / "libc.a";
	auto result = obj;
	result.replace_extension("");
	cmd = std::string("ld -m elf_i386 -L'") + lib.string() + "' '" + runtime.string() + "' '" + libc.string() + "' '" + obj.string() + "' -o '" + result.string() + "'";
	cerr << cmd << endl;
	system(cmd.data());
}

string Compiler::compile(Visitable *v) {
	buf <<
		".globl _start\n"
		".globl exit\n"
		"_start:\n"
		"	pushl %ebp\n"
		"	movl %esp, %ebp\n"
		"	call main\n"
		"	pushl %eax\n"
		"	call exit\n";

	for (const auto& c : scope->classes) {
		c->calculateSize();
	}

	for (const auto& f : scope->functions) {
		compileFunction(f);
	}

	for (const auto& it : quadGen.getStrings()) {
		buf << "_str_" << it.second << ":\n"
			<< ".string " << std::quoted(it.first) << "\n";
	}

	return buf.str();
}

void Compiler::compileFunction(const FunctionInfoPtr& f) {
	string name = mangleFunctionName(f);
	buf << ".globl " << quoted(name) << "\n";

	if (!f->block) return;

	// TODO: jeśli funkcja jest klasowa, ustawić self na ... wartosc z ecx
	// c++ thiscall ecx, stack right-to-left (push c, push b, push a)

	buf << quoted(name) << ":\n";
	stringstream asmGen;
	swap(buf, asmGen);
	exitLabel = make_shared<QLabel>("." + name + "_exit");
	QuadrupleGenerator::Result quadEnv = quadGen.compileFunction(f);
	try {
		RegisterAllocator regGen((quadEnv));
		regGen.allocate();
	} catch(const ParseError &e) {
		printFunction(quadEnv, f);
		throw ParseError(buf.str() + e.what());
	}

	printFunction(quadEnv, f);
	swap(buf, asmGen);
	uint localsCount = temps.size() + localInfos.size();
	for (auto i : Register::calleeSaved)
		append("PUSH", i);
	buf <<
	    "\tpush %ebp\n"
	    "\tmov %esp, %ebp\n";
	if (localsCount) {
		buf << "\tsub $" + to_string(localsCount * 4) << ",%esp\n";
	}
	buf << asmGen.str();
	buf << quoted(exitLabel->label) << ":\n" <<
		"\tMOVL %ebp, %esp\n"
	    "\tPOP %ebp\n";
	for (auto i = Register::calleeSaved.rbegin(); i != Register::calleeSaved.rend(); i++)
		append("POP", *i);
	buf << "\tRET\n";

	for (auto &t : temps) t->localOffset = 0;
	for (auto &v : localInfos) v->localOffset = 0;
	temps.clear();
	localInfos.clear();
}

void Compiler::printFunction(const QuadrupleGenerator::Result& quadEnv, const FunctionInfoPtr& f) {
	int blkNo = 0;
	buf << COMMENT "\x1b[32;1mFunction " << mangleFunctionName(f) << "\x1b[0m" << endl;
	for (const auto& b : quadEnv.blocks) {
		buf << COMMENT "----------------------------------------------------------------------------\n";
		buf << COMMENT "blok " << blkNo << " " << b->getName() << endl;
		/*buf << "in: ";
		for (auto in : b->in) {
			buf << in->getName() << " ";
		}
		buf << "\nouts: ";
		for (auto buf : b->buf) {
			buf << buf->getName() << " ";
		}
		buf << "\n";*/
		if (!b->flow.in.empty()) {
			buf << COMMENT "in: ";
			for (const auto& v : b->flow.in) {
				buf << v->name << ", ";
			}
			buf << "\n" COMMENT "----------\n";
		}
		if (!b->flow.def.empty()) {
			buf << COMMENT "def: ";
			for (const auto& v : b->flow.def) {
				buf << v->name << ", ";
			}
			buf << "\n" COMMENT "----------\n";
		}
		for (const auto& q : b->quads) {
			auto ret = q->toString();
			if (!ret.empty()) {
				auto len = 0;
				for (const auto &c : ret) {
					len++;
					if (c == '\t') len += 8 - (len % 8);
				}
				for (; len < 70; len++) ret += ' ';
				buf << COMMENT << ret << " alive: ";
				for (const auto &var : q->aliveAfter)
					buf << var->name << " ";
				buf << endl;

				stringstream asmStr;
				swap(asmStr, buf);
				q->generateAsm(*this);
				swap(asmStr, buf);

				if (!asmStr.str().empty()) {
					buf << COMMENT "\x1b[37;1m\n"
						<< asmStr.str()
						<< COMMENT "\x1b[0m\n";
				}
			}
		}
		if (!b->flow.use.empty()) {
			buf << COMMENT "----------\n" COMMENT "use: ";
			for (const auto &v : b->flow.use) {
				buf << v->name << ", ";
			}
			buf << "\n";
		}
		if (!b->flow.out.empty()) {
			buf << COMMENT "----------\n" COMMENT "out: ";
			for (const auto &v : b->flow.out) {
				buf << v->name << ", ";
			}
			buf << "\n";
		}
		blkNo++;
	}
}

const char* const Register::names[9] = {"eax", "ebx", "ecx", "edx", "edi", "esi", "ebp", "esp", "eip"};
const vector<Register> Register::all = {0, 1, 2, 3, 4, 5, 6, 7, 8};
const vector<Register> Register::callerSaved = {2, 3};
const vector<Register> Register::calleeSaved = {1, 4, 5};
const vector<Register> Register::assignable = {1,4,5,2,3}; // ecx and edx least used

string Compiler::getRef(const VariablePtr &v) {
	if (v->constExpr) {
		Str s;
		if (v->type->isEqual(&s) && v->val != 0) {
			return "$_str_" + to_string(v->val);
		}
		return "$" + to_string(v->val);
	}
	if (v->registerColor == 0) {
		if (!v->localOffset) {
			if (v->info) {
				if (!v->info->localOffset) {
					localInfos.emplace(v->info);
					v->info->localOffset = 4 * (temps.size() + localInfos.size());
				}
				v->localOffset = v->info->localOffset;
			} else {
				temps.emplace(v);
				v->localOffset = 4 * (temps.size() + localInfos.size());
			}
			if (v->localOffset > 0) v->name += "+";
			v->name += to_string(v->localOffset);
		}
		return to_string(-v->localOffset) + "(%ebp)";
	} else {
		return Register(v);
	}
}

void Compiler::generateQAlloc(QAlloc &q) {
	// memory = extension
	throw runtime_error("Memory alloc unimplemented!");
}

void Compiler::generateQWrite(QWrite &q) {
	// memory = extension
	throw runtime_error("Memory write unimplemented!");
}

void Compiler::generateQAccess(QAccess &q) {
	// memory = extension
	throw runtime_error("Memory access unimplemented!");
}

void Compiler::generateQReturn(QReturn &q) {
	if (q.val)
		moveTo(q.val, 0);

	append("JMP", quote(exitLabel->label));
}

void Compiler::generateQCall(QCall &q) {
	if (q.self) {
		throw runtime_error("Class call unimplemented!");
	}
	append("CALL", quote(mangleFunctionName(q.fn)));
	if (q.params) {
		append("ADDL", "$" + to_string(q.params*4), "%esp");
	}
	auto popRegs = registersToSave(q);
	for (auto r = popRegs.rbegin(); r != popRegs.rend(); r++) {
		if (*r != Register(q.loc))
			append("POPL", *r);
	}
	if (q.aliveAfter.count(q.loc)) {
		append("MOVL", "%eax", getRef(q.loc));
	}
}

void Compiler::generateQParam(QParam &q) {
//	return Quadruple::generateAsm();
	if (q.num < 0) {
		auto reg = Register(q.param);
		auto offset = 4 * (q.num - Register::calleeSaved.size() - 1); // push regs + ebp
		if (reg == 0) {
			q.param->localOffset = offset;
			if (q.param->info) q.param->info->localOffset = offset;
		} else {
			append("MOVL", to_string(-offset)+"(%ebp)", reg);
		}
	} else {
		if (q.num == 1 && q.call) {
			// first argument, need to save registers
			for (auto r : registersToSave(*q.call)) {
				if (r != Register(q.call->loc))
					append("PUSHL", r);
			}
		}
		append("PUSHL", getRef(q.param));
	}
}

void Compiler::generatePhiXCHG(const map<VariablePtr, VariablePtr> &varMap) {
	vector<Register> regMap(Register::all.size(), 0);
	// save to memory
	for (const auto &phiVars : varMap) {
		const auto &loc = phiVars.first, &val = phiVars.second;
		auto ri = Register(loc), rj = Register(val);
		if (loc->constExpr) continue;
		if (ri == 0) { // we need to do do memory stores now
			if (ri == 0 && rj == 0 && !val->constExpr) {
				assert(loc->info == val->info);
			} else {
				append("MOVL", getRef(val), getRef(loc));
			}
			continue;
		}
		auto &r = regMap[(int)ri];
		assert(r == 0 || r == rj);
		r = rj;
	}
	// remap registers
	for (uint x = 1; x < regMap.size(); x++) {
		auto ri = Register(x);
		auto &rj = regMap[x];
		if (rj == 0 || ri == rj) continue;
		int riUses = 0, rjUses = 0;
		for (auto r : regMap) {
			if (r == ri) riUses++;
			if (r == rj) rjUses++;
		}
		if (riUses) {
			append("XCHG", ri, rj);
			for (auto &r : regMap) {
				if (r == ri) r = rj;
				else if (r == rj) r = ri;
			}
		} else {
			append("MOVL", rj, ri);
			rj = 0;
		}
	}
	// load from memory
	for (const auto &phiVars : varMap) {
		const auto &i = phiVars.first, &j = phiVars.second;
		auto ri = Register(i), rj = Register(j);
		if (ri != 0 && (rj == 0 || j->constExpr)) {
			append("MOVL", getRef(j), ri);
		}
	}
}


void Compiler::generateQJump(QJump &q) {
	auto thisBlock = q.block;
	auto tgBlock = q.target->block;
	const auto &map = tgBlock->getPhiMapping(thisBlock);
	generatePhiXCHG(map);
	append("JMP", getBlockLabelText(q.target));
}

Op Compiler::generateCmp(const VariablePtr& l, Op op, const VariablePtr &r) {
	auto regL = Register(l), regR = Register(r);
	auto locL = getRef(l);
	auto locR = getRef(r);
	if (regL == 0 && regR == 0) {
		append("MOVL", locR, regR);
		locR = regR;
	} else if (regL != 0 && regR == 0) {
		// left is in register but right is not - swap operands
		//  as right needs to be a register
		swap(locL, locR);
		assert (op.kind() == Op::CMP);
		op = op.rot(); // we swapped operands
	}
	append("CMP", locL, locR);
	if (op.kind() == Op::CMP)
		op = op.rot(); // compare has different syntax in AT&T
	return op;
}

void Compiler::generateQJumpCond(QJumpCond &q) {
	auto thisBlock = q.block;
	auto tgBlock = q.target->block;
	const auto &map = tgBlock->getPhiMapping(thisBlock);
	for (const auto &phiVars : map) {
		if ((phiVars.first->info && phiVars.second->info && phiVars.first->info != phiVars.second->info) ||
			(phiVars.first->localOffset && phiVars.second->localOffset && phiVars.first->localOffset != phiVars.second->localOffset) ||
			(phiVars.first->registerColor && phiVars.second->registerColor && phiVars.first->registerColor != phiVars.second->registerColor)
			) {
			buf << ("// Unimplemented phi CondJump register XCHG\n");
		}
	}
	auto op = q.op;
	if (q.left) { // 2 args
		assert(!q.left->constExpr || !q.right->constExpr);
		op = generateCmp(q.left, op, q.right);
	} else { // 1 arg
		if (q.right->constExpr) {
			if ((op == Op::Not) ^ q.right->val) {
				generateQJump(q);
			}
			return;
		}
		auto tg = moveToAnyRegister(q.right);
		append("CMP", "$1", tg);
	}
	if (op.kind() != Op::CMP && op.kind() != Op::UNARY) throw runtime_error("Unexpected operator in compare");
	string operand = getOpName(op);
	// getOpName returns SETZ, SETNE, SET*, replace it to JMP
	operand.replace(0,3, "J");
	append(operand, getBlockLabelText(q.target));
}

void Compiler::generateQAssign(QAssign &q) {
	if (!q.aliveAfter.count(q.loc) || q.loc->constExpr) return;
	Register tg(q.loc);
	auto op = q.op;
	if (q.args.size() == 1) {
		auto &arg = q.args[0];
		switch (op.op) {
			case Op::Not: {
				auto reg = Register(arg);
				if (reg == 0) {
					// avoids extra MOV when arg is already in register
					moveTo(arg, tg);
					reg = tg;
				}
				append("CMP", "$1", reg);
				break;
			}
			case Op::Copy: {
				if (arg->constExpr) {
					append("MOVL", "$"+to_string(arg->val), getRef(q.loc));
					return;
				} else {
					moveTo(arg, tg);
					if (tg == 0) {
						append("MOVL", tg, getRef(q.loc));
					}
					return;
				}
				break;
			}
			case Op::Neg: {
				moveTo(arg, tg);
				append("NEG", tg);
				break;
			}
			default: throw runtime_error("Unexpected operand");
		}
	} else {
		auto &i = q.args[0], &j = q.args[1];
		if (op == Op::Div || op == Op::Mod) {
			auto edx = Register("edx");

			auto alives = to_set(aliveAfter(q, Register::assignable));

			bool edxAlive = alives.count(edx);

			bool backupEdx = edx != tg && edx != Register(i) && (edx == Register(j) || edxAlive);
			if (backupEdx) append("MOVL", edx, "-4(%esp)");
			string jLoc = Register(j) != edx ? getRef(j) : "-4(%esp)";
			if (j->constExpr) {
				auto free = to_set(Register::assignable) - alives;
				free.erase(Register("edx"));
				if (!free.empty()) {
					auto reg = *free.begin();
					append("MOVL", jLoc, reg);
					jLoc = reg;
				} else {
					append("MOVL", jLoc, "-8(%esp)");
					jLoc = "-8(%esp)";
				}
			}
			moveTo(i, 0);
			append("CLTD"); // edx lost
			append("IDIVL", jLoc);
			append("MOVL", op == Op::Mod ? edx : 0, getRef(q.loc));
			// restore unnecessary if j was in edx and becomes dead and was NOT coalesced
			// not easy to check if not coalesced, so move anyway
			if (backupEdx && edxAlive) append("MOVL", "-4(%esp)", edx);
			return;
		} else if (op.kind() == Op::CMP) {
			op = generateCmp(i, op, j);
		} else {
			assert(op.kind() == Op::BINARY);
			bool swapped = false;
			if (Register(j) == tg && tg != 0) {
				swapped = true;
				swap(i, j);
			}
			moveTo(i, tg);
			append(op, getRef(j), tg);
			if (swapped) {
				if (op == Op::Minus)
					append("NEG", tg);
			}
		}
	}
	if (op.boolean()) {
		if ((int) tg < 4) { // EAX - EDX
			append("MOVL", "$0", tg);
			string s = "%?l";
			s[1] = ((string)tg)[2];
			append(op, s);
		} else { // EDI/ESI
			append("MOVL", "$0", "%eax");
			append(op, "%al");
			append("MOVL", "%eax", tg);
		}
	}
	if (tg == 0) {
		append("MOVL", tg, getRef(q.loc));
	}
}

void Compiler::generateQLabel(QLabel &q) {
	buf << quoted(q.label) << ":" << endl;
}

void Compiler::generateQPhi(QPhi &q) {

}

void Compiler::generateQBlockEnd(QBlockEnd &q) {
	auto thisBlock = q.block;
	auto tgBlock = q.next;
	const auto &map = tgBlock->getPhiMapping(thisBlock);
	generatePhiXCHG(map);
}
