#ifndef COMPILER_HEADER
#define COMPILER_HEADER

#include "TypeDefs.h"
#include "Absyn.h"
#include "Skeleton.h"
#include <filesystem>
#include <utility>
#include "codeGen/Variable.h"
#include "ParseError.h"
#include "Info.h"
#include "codeGen/Quadruple.h"
#include "codeGen/BasicBlock.h"
#include "codeGen/QuadrupleGenerator.h"
#include <unordered_set>

class Register {
public:
	Register(string n) {
		for (auto &reg : names) {
			if (n == reg) {
				i = &reg - names;
				return;
			}
		}
		throw runtime_error("Invalid register name");
	};
	Register(int n) : i(n) { }
	Register(const VariablePtr &v) : i(v->registerColor) { }
	int i;
	operator string() const { return string("%") + names[i]; }
	explicit operator int() const { return i; }
	bool operator==(const Register &other) const { return i == other.i; }
	bool operator!=(const Register &other) const { return i != other.i; }
	bool operator<(const Register &other) const { return i < other.i; }

//	static const Register &eax, &ebx, &ecx, &edx, &edi, &esi, &ebp, &esp;
	static const char* const names[9];
	static const vector<Register> all;
	static const vector<Register> callerSaved;
	static const vector<Register> calleeSaved;
	static const vector<Register> assignable;
};

using Op = Quadruple::Op;
class Compiler {
public:
	Compiler(const std::string& f, shared_ptr<Scope> s) : file(f), buf(), scope(s), quadGen(std::move(s)) {};

	static const std::string extension;
	static void externalCommand(std::filesystem::path file);

	string compile(Visitable *v);

	static std::string getVirtName(const ClassInfoPtr& c) {
		return c->name + ":virt_table";
	}

	void generateQJumpCond(QJumpCond &q);
	void generateQAlloc(QAlloc &q);
	void generateQWrite(QWrite &q);
	void generateQAccess(QAccess &q);
	void generateQReturn(QReturn &q);
	void generateQCall(QCall &q);
	void generateQParam(QParam &q);
	void generateQJump(QJump &q);
	void generateQAssign(QAssign &q);
	void generateQLabel(QLabel &q);
	void generateQPhi(QPhi &q);
	void generateQBlockEnd(QBlockEnd &q);

	Op generateCmp(const VariablePtr& l, Op op, const VariablePtr &r);

	static std::string mangleFunctionName(const FunctionInfoPtr& f) {
		if (auto c = f->klass.lock()) {
			return c->name + "::" + f->name;
		}
		return f->name;
	}
private:
	std::filesystem::path file;
	std::stringstream buf;
	shared_ptr<Scope> scope;
	QuadrupleGenerator quadGen;

	unordered_set<VariablePtr> temps;
	unordered_set<VarInfoPtr> localInfos;

	void compileFunction(const FunctionInfoPtr& f);
	void printFunction(const QuadrupleGenerator::Result& quadEnv, const FunctionInfoPtr& f);

	string getRef(const VariablePtr &v);
	void generatePhiXCHG(const map<VariablePtr, VariablePtr> &varMap);

	static string getOpName(Quadruple::Op op) {
		switch(op.op) {
			case Op::Plus: return "ADDL";
			case Op::Minus: return "SUBL";
			case Op::Mul: return "IMULL";
			case Op::Div:
			case Op::Mod: throw runtime_error("Impossible to handle division with default");
			case Op::Not: return "SETNZ";
			case Op::Copy: return "SETZ";
			case Op::LT: return "SETL";
			case Op::LE: return "SETLE";
			case Op::EQ: return "SETZ";
			case Op::NEQ: return "SETNZ";
			case Op::GE: return "SETGE";
			case Op::GT: return "SETG";
			default: return "Op" + to_string(op.op) + "?";
		}
	}

	void moveTo(VariablePtr &v, Register tg) {
		if (v->constExpr) {
			append("MOVL", "$"+to_string(v->val), tg);
			return;
		}
		auto reg = Register(v);
		if (reg == 0) {
			append("MOVL", getRef(v), tg);
		} else if (reg != tg) {
			append("MOVL", reg, tg);
		}
	}

	void moveTo(Register reg, Register tg) {
		if (reg != tg) {
			append("MOVL", reg, tg);
		}
	}

	Register moveToAnyRegister(VariablePtr &v) {
		auto reg = Register(v);
		if (reg == 0) {
			append("MOVL", getRef(v), reg);
		}
		return reg;
	}

	void append(Quadruple::Op op, string op1, string op2 = "", string op3 = "") {
		return append(getOpName(op), op1, op2, op3);
	}

	void append(string cmd, string op1 = "", string op2 = "", string op3 = "") {
		append({cmd, op1, op2, op3});
	}

	void append(vector<string> ops) {
		static const char* const delimiters = "\t ,,,,";
		assert(ops.size() < 6);
		while (ops.size() && ops.back().empty())
			ops.pop_back();
		buf << "\t";
		for (uint i = 0; i < ops.size(); i++) {
			buf << delimiters[i] << ops[i];
		}
		buf << endl;
	}


	string getBlockLabelText(const QuadruplePtr &q) {
		return getBlockLabelText(q->block);
	}

	string getBlockLabelText(BasicBlockPtr &b) {
		auto n = b->getName();
		if (n.empty()) throw runtime_error("Attempt to get label text of unlabeled block");
		return quote(n);
	}

	shared_ptr<QLabel> exitLabel;
	void generateStandardReturn() {
		append("JMP", quote(exitLabel->label));
	}

	static vector<Register> aliveAfter(Quadruple &q, const vector<Register> &vec) {
		const uint noRegs = vec.size();
		bool used[noRegs];
		int notUsed = noRegs;
		for (auto &i : used) i = false;
		for (auto &live : q.aliveAfter) {
			for (uint i = 0; i < noRegs; i++) {
				const Register &r = vec[i];
				if (!used[i] && Register(live) == r) {
					used[i] = true;
					--notUsed;
				}
			}
			if (notUsed == 0) break;
		}
		vector<Register> ret;
		for (uint i = 0; i < noRegs; i++) {
			if (used[i]) ret.emplace_back(vec[i]);
		}
		return ret;
	}
	static vector<Register> registersToSave(QCall &q) {
		return aliveAfter(q, Register::callerSaved);
	}
};

#endif
