#ifndef ZAD2_GRAPH_H
#define ZAD2_GRAPH_H
#include <unordered_set>
#include <queue>
using namespace std;

class FindUnion {
	uint n{0};
	vector<uint> t;
public:
	void resize(uint num) {
		n = num;
		t.resize(n);
		for (uint i = 0; i < n; i++)
			t[i] = i;
	}

	uint get(uint c) {
		if (t[c] == c) return c;
		return t[c] = get(t[c]);
	}

	void add(uint a, uint b) {
		t[get(a)] = get(b);
	}
};

template<typename T>
class Graph {
public:
	void add(const T &i) {
		if (evaluated) throw runtime_error("Add to evaluated graph");
		vertices[i];
	}

	void add(const T &i, const T &j) {
		if (evaluated) throw runtime_error("Add to evaluated graph");
		vertices[i];
		vertices[j];
		edgesSet.emplace(i, j);
		edgesSet.emplace(j, i);
	}

	void addAssign(const T &i, const T &j) {
		if (evaluated) throw runtime_error("Add to evaluated graph");
		assignsSet.emplace(i, j);
		vertices[i];
		vertices[j];
	}

	const auto& getEdges() const {
		return edgesSet;
	}

	const auto& getAssigns() const {
		return assignsSet;
	}
	map<T, int> getAllocation() const {
		if (!evaluated) throw runtime_error("Query of unevaluated graph");
		map<T, int> ret;
		for (const auto &i : vertices)
			ret[i.first] = color[i.second];
		return ret;
	}

	int getColor(T x) {
		if (!evaluated) throw runtime_error("Query of unevaluated graph");
		if (!vertices.count(x)) throw runtime_error("Query of non-graph vertex");
		int c = vertices[x];
		return color[c];
	}

	void eval(int colors) {
		if (evaluated) throw runtime_error("Try of re-evaluating graph");
		evaluated = true;
		if (vertices.empty()) return;
		maxColors = colors;
		buildGraph();
		initOrder();
		initChordal();
		initColoring(vertices.size());
		if (!chordal) return;
		if (maxColors < 1) {
			throw std::runtime_error("Spill to no colors!");
		}
		spillToColors(maxColors);
		coalesce();
		remapColors();
		checkAllocation();
	}

	bool isChordal() const {
		if (!evaluated) throw runtime_error("Query of unevaluated graph");
		return chordal;
	}
private:
	bool evaluated = false;
	set<pair<T, T>> edgesSet;
	map<T, int> vertices;
	vector<T> reverseVertices;
	vector<unordered_set<int>> edges;
	FindUnion alias;
	vector<int> order;
	vector<int> color;
	bool chordal = true;
	int maxColors = 0;
	set<pair<T, T>> assignsSet;

	void buildGraph() {
		int n = 0;
		reverseVertices.reserve(vertices.size());
		for (auto &i : vertices) {
			i.second = n++;
			reverseVertices.emplace_back(i.first);
		}

		alias.resize(vertices.size());
		edges.resize(vertices.size());
		for (auto e : edgesSet) {
			auto a = vertices[e.first], b = vertices[e.second];
			edges[a].emplace(b);
		}
	}

	void initOrder() {
		if (vertices.empty()) return;
		vector<int> card;
		vector<unordered_set<int>> lambda;
		int n = vertices.size();
		card.resize(n);
		lambda.resize(n);
		for (int i = 0; i < n; i++) {
			lambda[0].emplace(i);
		}

		while (!lambda.empty()) {
			if (lambda.back().empty()) {
				lambda.pop_back();
				continue;
			}
			auto v = *lambda.back().begin();
			order.push_back(v);
			for (auto x : edges[v]) {
				auto &c = card[x];
				if (c >= 0) {
					lambda[c].erase(x);
					c++;
					if ((int)lambda.size() <= c)
						lambda.emplace_back();
					lambda[c].emplace(x);
				}
			}
			lambda[card[v]].erase(v);
			card[v] = -1;
		}
	}

	void initChordal() {
		unordered_set<int> visited;
		for (auto i : order) {
			unordered_set<int> neighbors;
			for (auto x : edges[i]) {
				if (visited.count(x)) {
					neighbors.emplace(x);
				}
			}
			for (auto x : neighbors) {
				for (auto y : neighbors) {
					if (x == y) continue;
					if (!edges[x].count(y) || !edges[y].count(x)) {
						chordal = false;
						return;
					}
				}
			}
			visited.emplace(i);
		}
		chordal = true;
	}

	vector<int> spilledVertices;
	int highestColor = -1;
	vector<unordered_set<int>> colorUses;
	vector<int> colorCount;
	vector<int> spilledColor;
	vector<uint> nonSpilledColors;
	vector<int> coalescedNodes;

	void initColoring(int maxC) {
		highestColor = -1;
		color.resize(vertices.size());
		colorUses.resize(vertices.size());
		colorCount.resize(vertices.size());
		vector<char> used;
		used.resize(maxC);
		for (auto &i : color) i = -1;
		for (auto c : order) {
			// calc used
			for (auto x : edges[c]) {
				if (color[x] == -1) continue;
				used[color[x]] = true;
			}

			// calc self color
			for (int col = 0; col < maxC; col++) {
				if (!used[col]) {
					color[c] = col;
					break;
				}
			}

			// clear used table for next use
			for (auto x : edges[c]) {
				if (color[x] == -1) continue;
				used[color[x]] = false;
			}

			// spill or apply color
			if (color[c] == -1) {
				markSpilled(c);
			} else {
				highestColor = max(highestColor, color[c]);
				colorUses[color[c]].emplace(c);
				colorCount[color[c]]++;
			}
		}
		while (!colorCount.empty() && colorCount.back() == 0) {
			colorUses.pop_back();
			colorCount.pop_back();
		}
	}

	bool spilled(int c) const {
		return color[c] == -1;
	}

	void markSpilled(int c) {
		spilledVertices.emplace_back(c);
		color[c] = -1;
		// edges won't contain already spilled nodes
		for (auto x : edges[c]) {
			edges[x].erase(c);
		}
	}

	void spillToColors(uint maxC) {
		// heuristic: least used
		priority_queue<pair<uint, uint>> Q; // { number of uses, color id }
		spilledColor.resize(colorCount.size());
		nonSpilledColors.clear();
		for (uint i = 0; i < colorCount.size(); i++) {
			Q.emplace(colorCount[i], i);
		}

		for (uint i = 0; i < maxC && !Q.empty(); i++) {
			// pop first most used maxC colors
			auto col = Q.top().second;
			spilledColor[col] = false;
			nonSpilledColors.emplace_back(col);
			Q.pop();
		}

		while (!Q.empty()) {
			auto col = Q.top().second;
			spilledColor[col] = true;
			Q.pop();
			for (auto i : colorUses[col]) {
				if (!spilled(i)) {
					markSpilled(i);
				}
			}
			colorUses[col].clear();
		}

		// reconnect spilled nodes;
		for (auto c : spilledVertices) {
			for (auto it = edges[c].begin(); it != edges[c].end();) {
				auto n = *it;
				if (spilled(n)) {
					edges[n].emplace(c);
					++it;
				} else {
					it = edges[c].erase(it);
				}
			}
		}
	}

	void setColor(int x, int col) {
		if (color[x] == col) return;
		if (color[x] != -1) {
			colorUses[color[x]].erase(x);
			colorCount[color[x]]--;
		}
		color[x] = col;
		if (col != -1) {
			colorCount[col]++;
			colorUses[col].emplace(x);
		}
	}

	void coalesceNode(int x, int y, int col) {
		for (auto e : edges[y]) {
			edges[x].emplace(e);
			edges[e].emplace(x);
			edges[e].erase(y);
		}
		edges[y].clear();
		setColor(x, col);
		coalescedNodes.emplace_back(y);
		alias.add(y, x);
	}

	void coalesce() {
		vector<char> used;
		used.resize(highestColor + 1);
		for (auto it : assignsSet) {
			for (auto &c : used) c = false;
			auto x = vertices[it.first], y = vertices[it.second];
			x = alias.get(x); y = alias.get(y);
			if (x == y) continue;
			if (spilled(x) || spilled(y)) continue;
			if (edges[x].count(y) || edges[y].count(x)) continue;
			for (auto z : edges[x]) if (color[z] >= 0) used[color[z]] = true;
			for (auto z : edges[y]) if (color[z] >= 0) used[color[z]] = true;
			for (int c : nonSpilledColors) {
				if (used[c]) continue;
				coalesceNode(x, y, c);
				break;
			}
		}
		for (uint i = 0; i < color.size(); i++) {
			setColor(i, color[alias.get(i)]);
		}
	}

	void remapColors() {
		sort(nonSpilledColors.begin(), nonSpilledColors.end());
		for (uint i = 0; i < nonSpilledColors.size(); i++) {
			if (nonSpilledColors[i] != i) {
				auto &col = nonSpilledColors[i];
				for (auto v : colorUses[col])
					color[v] = i;
				swap(colorCount[i], colorCount[col]);
				swap(colorUses[i], colorUses[col]);
			}
		}
	}

	void checkAllocation() {
		if (vertices.size() != color.size()) throw runtime_error("Invalid allocation, vertices size mismatches color size");
		uint size = spilledVertices.size() + coalescedNodes.size();
		for (uint i = 0; i < vertices.size(); i++) if (!spilled(i)) if (alias.get(i) == i) size++;
		if (size != vertices.size()) throw runtime_error("Invalid allocation, original size mismatches current size");
		for (auto v : edgesSet) {
			int x = alias.get(vertices[v.first]);
			int y = alias.get(vertices[v.second]);
			if (!spilled(x) && color[x] == color[y])
				throw runtime_error("Invalid allocation, same color assigned to colliding variables");
			if (color[x] >= maxColors || color[y] >= maxColors)
				throw runtime_error("Invalid allocation, more colors used than allowed");
		}
	}
};

#endif //ZAD2_GRAPH_H
