#include "RegisterAllocator.h"
#include "BasicBlock.h"
#include "Util.h"
#include <fstream>
#include "../Compiler.h"

void RegisterAllocator::analyseLive() {
	for (const auto& v : vars) {
		for (const auto& q : v->writes) {
			q->block->flow.addDef(v);
		}
		if (v->constExpr) continue;
		for (const auto& q : v->uses) {
			q->block->flow.addUse(v);
		}
	}

	for (const auto& b : blocks) {
		for (const auto& phiMap : b->phi->phi) {
			for (const auto& phiRow : phiMap) {
				b->flow.def.emplace(phiRow.first);
			}
		}
	}

	while (true) {
		bool changed = false;
		for (const auto& b : blocks) {
			set<VariablePtr> pIn, pOut, diff;
			auto &f = b->flow;
			pIn.swap(f.in);
			pOut.swap(f.out);
			f.in = f.use + (pOut - f.def);
			for (const auto& succ : b->out) {
				f.out += succ->flow.inForBlock(b);
			}
			if (f.in != pIn || f.out != pOut) changed = true;
		}
		if (!changed) break;
	}

	if (!blocks.empty() && !blocks[0]->flow.in.empty())
		throw ParseError("Variable definition not found - first block expects input");
}

void RegisterAllocator::allocate() {
	analyseLive();
	buildGraph();
}

void printGraph(const Graph<VariablePtr> &graph, const std::string& filename) {
	ofstream f(filename);
	for (const auto &v : graph.getEdges()) {
		f << v.first->name << " " << v.second->name << endl;
	}
	f.close();
}

void printGraph(const Graph<std::string> &graph, const std::string& filename) {
	ofstream f(filename);
	for (const auto &v : graph.getEdges()) {
		f << v.first << " " << v.second << endl;
	}
	f.close();
}

void RegisterAllocator::buildGraph() {
	for (const auto& b : blocks) {
		set<VariablePtr> alive = b->flow.in;
		for (const auto& q : b->quads) {
			alive += to_set(q->definitions());
			for (const auto& al : alive) assert(al);
			vector<VariablePtr> dead;
			for (const auto& var : alive) {
				// this quad is the last quad that uses var
				auto noMoreUses = var->uses.empty() || var->uses.back() == q;
				if ((noMoreUses && !b->flow.out.count(var)) || var->constExpr) {
					dead.emplace_back(var);
				}
			}
			alive -= to_set(dead);
			q->aliveAfter = alive;
			for (auto i = alive.begin(); i != alive.end(); i++) {
				graph.add(*i);
				auto j = i;
				for (++j; j != alive.end(); ++j) {
					graph.add(*i, *j);
				}
			}
			if (auto assign = dynamic_pointer_cast<QAssign>(q)) {
				if (assign->op == Quadruple::Op::Copy && !assign->loc->constExpr && alive.count(assign->loc)) {
					assert(assign->loc);
					assert(assign->args[0]);
					graph.addAssign(assign->loc, assign->args[0]);
				}
			} else if (auto phi = dynamic_pointer_cast<QPhi>(q)) {
				for (const auto &phiBlock : phi->phi) {
					for (const auto &phiVars : phiBlock) {
						if (alive.count(phiVars.first)) {
							assert(phiVars.first);
							assert(phiVars.second);
							graph.addAssign(phiVars.first, phiVars.second);
						}
					}
				}
			}
		}
	}
	graph.eval(Register::assignable.size());
	int spills = 0;
	for (auto &v : graph.getAllocation()) {
		if (v.second == -1) {
			v.first->registerColor = 0;
			spills++;
		} else {
			auto regNo = (int)Register::assignable[v.second];
			v.first->registerColor = regNo;
			v.first->name += "-r" + to_string(regNo + 1);
		}
	}
	if (!graph.isChordal()) {
//		printGraph(graph, "./mygraph");
		throw ParseError("allocation: graph not chordal!");
	}
}


